#* Imports
import pycook.elisp as el
lf = el.lf

#* Functions
def elisp_files_cwd():
    return el.directory_files(el.default_directory(), False, ".*el$")

#* Recipes
def install(recipe, *packages):
    packages_str = " ".join(packages)
    return el.emacs_batch_eval(lf("(cook-install-deps '({packages_str}))"))

def byte_compile(recipe, *fnames):
    fnames = fnames or elisp_files_cwd()
    fnames_str = " ".join(['"' + fname + '"' for fname in fnames])
    return el.emacs_batch_eval(lf("(cook-byte-compile {fnames_str})"))

def elpa(recipe, *fnames):
    """Start Emacs, load elpa.el, and FNAMES."""
    fnames = fnames or elisp_files_cwd()
    fnames = [el.emacs_cook_script("elpa.el"), *fnames]
    fnames_str = " ".join(['-l "' + fname + '"' for fname in fnames])
    return [
        "emacs --version",
        lf("emacs -Q {fnames_str}")
    ]



#* Imports
import os
import pycook.elisp as el
import pycook.insta as st
import pycook.recipes.git as git
import pycook.recipes.java as java
from datetime import datetime
lf = el.lf

def addpath(p):
    cp = os.getenv("CLASSPATH")
    if cp:
        ps = cp.split(":")
    else:
        ps = []
    if p in ps:
        return []
    else:
        ps.append(p)
        cp = ":".join(ps)
        return [lf("+export CLASSPATH=\"{cp}\"")]

#* Recipes
def ng_install(recipe):
    local = el.expand_file_name("~/git/clojure/nailgun/")
    st.git_clone("https://github.com/facebook/nailgun", local, "84f3b05")
    st.install_package("maven")
    jar = el.expand_file_name("nailgun-server/target/nailgun-server-1.0.0.jar", local)
    st.make(
        jar,
        [lf("cd {local}"), "mvn clean install"],
        [local])
    st.make(
        "/usr/local/bin/ng",
        [lf("cd {local}"), "sudo make install"])
    st.patch("~/.bashrc", addpath(jar))

def ng_server(recipe):
    return ["java com.facebook.nailgun.NGServer"]

def ng_clojure(recipe):
    return ["ng clojure.main"]



import pycook.elisp as el
import pycook.insta as st

def install(recipe, package=None):
    if recipe == 42:
        st.install_package(package)
    elif recipe[0] == "complete":
        return el.sc("apt-cache pkgnames " + recipe[1])

def history(recipe):
    return [
        """(zcat $(ls -tr /var/log/apt/history.log*.gz); cat /var/log/apt/history.log) 2>/dev/null |
            egrep '^(Start-Date:|Commandline:)' |
            grep -v aptdaemon |
            egrep '^Commandline:' | awk '!x[$0]++'
        """
    ]

def remove(recipe, package=None):
    if recipe == 42:
        return [el.lf("sudo apt-get remove {package}")]
    elif recipe[0] == "complete":
        return el.shell_command_to_string("dpkg --get-selections " + recipe[1] + "* | awk '{print $1}'")



#* Imports
import subprocess
import sys
import os
import re
import shlex
import collections
import traceback

#* Globals
sc_hookfn = None
cd_hookfn = None

#* Functional
def apply (function, arguments):
    """Call FUNCTION with ARGUMENTS, return the result."""
    return function (*arguments)

def mapcar (func, lst):
    """Compatibility function for Python3.

    In Python2 `map' returns a list, as expected.  But in Python3
    `map' returns a map object that can be converted to a list.
    """
    return list (map (func, lst))

def filter(pred, lst):
    return [x for x in lst if pred(x)]

cl_remove_if_not = filter

def cl_remove_if (pred, lst):
    return [x for x in lst if not pred(x)]

def cl_position_if (pred, lst):
    pos = 0
    for item in lst:
        if pred (item):
            return pos
        else:
            pos += 1

def position (item, lst, default = None):
    if item in lst:
        return lst.index(item)
    else:
        return default

def cl_set_difference(lst1, lst2):
    s = set(lst2)
    return [x for x in lst1 if x not in s]

def cl_find_if (pred, lst):
    for item in lst:
        if pred (item):
            return item

def position_if(pred, lst):
    for (i, item) in enumerate(lst):
        if pred(item):
            return i

def mapconcat (func, lst, sep):
    if func:
        return sep.join (map (func, lst))
    else:
        return sep.join (lst)

def print_last_error():
    (_, err, tb) = sys.exc_info()
    print(err)
    traceback.print_tb(tb)

def map_ignore_errors(func, items):
    """Evaluate FUNC for each item in ITEMS.
    Keep going with the next item on exception.
    Print the traceback for the first exception.
    """
    traceback_printed = False
    failed_items = []
    for item in items:
        try:
            func(item)
        except:
            failed_items.append(item)
            if not traceback_printed:
                print_last_error()
                traceback_printed = True
    return failed_items

def flatten (seq):
    """Flatten a list of lists into a list."""
    return [item for sublist in seq for item in sublist]

def partition (n, seq):
    return [seq[i:i + n] for i in range (0, len (seq), n)]

def delete (element, lst):
    return [x for x in lst if x != element]

def delete_dups(lst):
    seen = set()
    seen_add = seen.add
    return [x for x in lst if not (x in seen or seen_add(x))]

#* Sys
def top_level ():
    f = sys._getframe ()
    while f.f_back:
        f = f.f_back
    return f

def crash ():
    tf = top_level ()
    f = sys._getframe ().f_back
    tf.f_globals["lnames"] = f.f_locals.keys ()
    for (k, v) in f.f_locals.items ():
        tf.f_globals[k] = v
    raise RuntimeError ("locals stored to globals")

#* OS
def addpath (path):
    sys.path.append (path)

def user_login_name ():
    import getpass
    return getpass.getuser()

def emacsclient_eval(expr):
    e = re.sub('"', "\\\"", expr)
    return lf('emacsclient -e "{e}"')

def emacs_cook_script(fname):
    d_lib = locate_dominating_file(__file__, "lib")
    if d_lib:
        return expand_file_name("../cook/" + fname, d_lib)
    else:
        # sys.path has '', handle loading this package from git
        d_etc = locate_dominating_file(__file__, "etc")
        if fname == "cook.el":
            return expand_file_name("../" + fname, d_etc)
        else:
            return expand_file_name(fname, d_etc)

def emacs_batch_eval(expr):
    e = re.sub('"', "\\\"", expr)
    script_el = emacs_cook_script("scripts.el")
    return lf('emacs -batch -l {script_el} --eval "{e}"')

def eval (s):
    return shell_command_to_string(emacsclient_eval(s))

def beval(s, init_file = None):
    s = re.sub('"', "\\\"", s)
    if init_file:
        init = "-l "+ init_file
    else:
        init = ""
    return shell_command_to_string(lf('emacs -batch {init} --eval "(print {s})"'))

def load_file (f):
    "Load a Python file into the REPL."
    exec (open (f).read(), globals ())

#* Files
class dd:
    def __init__(self, d):
        self.d = expand_file_name(d)

    def __enter__(self):
        self._old_dir = default_directory()
        os.chdir(self.d)

    def __exit__(self, *_):
        os.chdir(self._old_dir)

def default_directory ():
    return os.getcwd ()

def locate_dominating_file (f, n):
    if file_directory_p(f):
        d = f
    else:
        d = file_name_directory(expand_file_name(f))
    while d != "/":
        nd = os.path.join(d, n)
        if file_exists_p(nd):
            return nd
        d = file_name_directory(d)

def cd (directory):
    d = expand_file_name (directory)
    if cd_hookfn:
        cd_hookfn(d)
    os.chdir (d)

def make_directory(d):
    """Work around Python2/3 `os.makedirs' incompat."""
    if not os.path.exists(d):
        os.makedirs(d)

def expand_file_name(f, directory = None):
    if not directory:
        directory = os.getcwd()
    else:
        directory = os.path.expanduser(directory)
    if re.match("^~", f):
        return os.path.expanduser (f)
    elif re.match("\.\./", f):
        return os.path.realpath(os.path.join(directory, f))
    else:
        return os.path.join(directory, f)

def path_join(*parts):
    return os.path.sep.join(parts)

def file_name_sans_extension(f):
    return os.path.splitext(f)[0]

def file_name_directory(f):
    return os.path.dirname(f)

def file_name_nondirectory(f):
    return os.path.basename(f)

def file_name(f):
    return file_name_sans_extension(file_name_nondirectory(f))

def file_exists_p (f):
    return os.path.exists (expand_file_name(f))

def file_newer_than_file_p(f1, f2):
    return os.path.getctime(f1) > os.path.getctime(f2)

def file_directory_p (f):
    return os.path.isdir (f)

def abbreviate_file_name (f, d):
    if not d[-1] == "/":
        d = d + "/"
    m = re.match (d, f)
    if m:
        return f[m.end ():]
    else:
        m = re.match (f, d)
        if m:
            return ".".join (["../"]* (d[m.end ():].count ("/") - 1))

def directory_files (d, full = False, match = False):
    fl = os.listdir (d)
    if match:
        fl = cl_remove_if_not (lambda f:  None != string_match (match, f), fl)
    if full:
        fl = mapcar (lambda f: expand_file_name (f, d), fl)
    return fl

def delete_file (f):
    return os.remove (f)

#* File read/write
def slurp (f):
    fh = open (f, 'r')
    res = fh.read ()
    fh.close ()
    return res

def slurp_lines(f):
    return open(expand_file_name(f), "r").read().splitlines()

def barf (f, s):
    fh = open (f, 'w')
    fh.write (s)
    fh.close ()

#* Shell
def shell_command_to_string(cmd, **kwargs):
    out = subprocess.check_output(["bash", "-c", cmd], **kwargs).strip()
    if type(out) is str:
        return out
    else:
        return out.decode()

def sc(cmd, **kwargs):
    fcmd = lf(cmd, 2)
    if sc_hookfn:
        sc_hookfn(fcmd)
    return shell_command_to_string(fcmd, **kwargs)

def shell_command_to_list(cmd, **kwargs):
    cmd_output = shell_command_to_string(cmd, **kwargs)
    return [s for s in cmd_output.split ("\n") if s]

def sc_l(cmd, **kwargs):
    fcmd = lf(cmd, 2)
    if sc_hookfn:
        sc_hookfn(fcmd)
    return shell_command_to_list(fcmd, **kwargs)

def bash(cmd, echo = False, capture = False, **kwargs):
    if type(cmd) is list:
        cmd = "\n".join(cmd)
    if echo:
        sep = "-"*80
        print(sep)
        print("Run: \n" + cmd)
        print(sep)
    sys.stdout.flush()
    if capture:
        p = subprocess.Popen(["/bin/bash", "-e", "-c", cmd], stdout=subprocess.PIPE, **kwargs)
        (stdout, stderr) = p.communicate()
        if p.returncode == 0:
            return stdout.decode().strip()
        else:
            raise subprocess.CalledProcessError(p.returncode, cmd, stdout, stderr)
    else:
        p = subprocess.Popen(["/bin/bash", "-e", "-c", cmd], **kwargs)
        return_code = p.wait()
        if return_code == 0:
            return 0
        else:
            sys.stdout.flush()
            sys.stderr.flush()
            raise subprocess.CalledProcessError(return_code, cmd)

#* String
def lf (string, lvl = 1):
    fr = sys._getframe()
    for i in range(lvl):
        fr = fr.f_back
    vars_dict = fr.f_globals.copy()
    vars_dict.update(fr.f_locals)
    return string.format(**vars_dict)

#* Regex
def string_match (regexp, string):
    global match_data
    m = re.search (regexp, string)
    if m:
        match_data = m
        return m.start ()

def match_string (group):
    global match_data
    return match_data.group (group)

def match_beginning (group):
    global match_data
    return match_data.start (group)

def match_end (group):
    global match_data
    return match_data.end (group)

def re_filter (regex, seq):
    return list (filter (lambda s: re.search (regex, s), seq))

def re_seq (regex, s):
    return re.findall (regex, s)

def re_match(regex):
    return lambda s: re.search(regex, s)

def replace_regexp_in_string (regexp, rep, string):
    return re.sub (re.compile(regexp, re.MULTILINE), rep, string)

def spit(x, fname):
    with open(fname, "w") as f:
        if type(x) is str:
            f.write(x)
        elif type(x) is list and type(x[0]) is str:
            if re.search("\n", x[0]):
                sep = ""
            else:
                sep = "\n"
            f.write(sep.join(x))
        else:
            raise RuntimeError("Unexpected object", x)

def re_split(regex, lst):
    l1 = []
    l2 = []
    for x in lst:
        if re.search(regex, x):
            l1.append(x)
        else:
            l2.append(x)
    return (l1, l2)

def group_by(f, lst):
    res = collections.OrderedDict()
    for x in lst:
        fx = f(x)
        if fx in res:
            res[fx].append(x)
        else:
            res[fx] = [x]
    return res

def re_extract(regex, group = 1):
    def res(s):
        m = re.search(regex, s)
        if m:
            return m.group(group)
    return res

#* Time
def timestamp():
    from datetime import datetime
    t = datetime.now()
    year = t.year
    month = t.month
    day = t.day
    hour = t.hour
    minute = t.minute
    dow = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"][t.weekday()]
    return lf("<{year}-{month:02d}-{day:02d} {dow} {hour:02d}:{minute:02d}>")



import pycook.elisp as el

def ls(recipe, package):
    if type(recipe) is int:
        return ["dpkg -L " + package]
    elif recipe[0] == "complete":
        part = recipe[1]
        if len(part) > 0:
            part += "*"
        return el.shell_command_to_string("dpkg-query -W -f '${Package}\n' " + part)


def file_to_package(recipe, fname):
    return ["dpkg -S " + fname]



#* Imports
import shlex

#* Recipes
def vlc_upload(recipe, fname):
    """VLC / Network / Sharing via WiFi."""
    return [
        "curl -F files[]=@" + shlex.quote(fname) + " http://192.168.178.12/upload.json"]



#* Imports
import pycook.elisp as el
import shlex

#* Recipes
def show_current_password(recipe):
    this_connection = el.sc_l("nmcli -t connection")[0]
    (name, *_) = this_connection.split(":")
    return [
        "sudo cat /etc/NetworkManager/system-connections/" + shlex.quote(name)
    ]



#* Imports
import pycook.elisp as el

#* Recipes
def burn_to_usb(recipe, fname):
    disks = el.sc_l("df")
    media_disks = el.re_filter("/media/", disks)
    assert len(media_disks) == 1
    usb_disk_data = media_disks[0].split()
    assert len(usb_disk_data) == 6
    usb_disk = usb_disk_data[0]
    fname_full = el.expand_file_name(fname)
    return [
        "sudo umount " + usb_disk,
        el.lf("sudo dd bs=4M if={fname_full} of={usb_disk} status=progress oflag=sync")]



#* Imports
import os
import re
import shlex
import pycook.elisp as el
lf = el.lf

#* Functions
def export_pdf(recipe, fname):
    fpdf = re.sub("org$", "pdf", fname)
    return [
        el.emacs_batch_eval(lf('(cs-org-to-pdf "{fname}")')),
        lf("nohup 1>/dev/null 2>/dev/null evince {fpdf} &"),
        el.emacsclient_eval('(with-current-buffer (window-buffer (selected-window)) (bury-buffer))')
    ]

def export_html(recipe, fname):
    fname = el.expand_file_name(fname)
    fhtml = re.sub("org$", "html", fname)
    el.sc(el.emacsclient_eval(lf('(cs-org-to-html "{fname}")')))
    if "INSIDE_EMACS" not in os.environ:
        return ["firefox " + shlex.quote(fhtml)]

#* Recipes
def clean(recipe):
    return ["rm -rf *.pdf *.tex _minted*"]



#* Imports
import sys
import shutil
import pycook.elisp as el
sc = el.sc
lf = el.lf

#* Constants
sudo_cmd = "sudo -H " if el.file_exists_p("/usr/bin/sudo") else "su -c"

#* Functions
def sudo(cmd):
    return sudo_cmd + cmd

def get_pip():
    is_sudo = shutil.which("cook").find("/usr/local/bin") != -1
    if sys.version_info.major == 3:
        pip = "pip3"
    else:
        pip = "pip"
    if is_sudo:
        return sudo(pip)
    else:
        return pip

def package_installed_p(package):
    try:
        s = sc(get_pip() + " show " + package)
        return s != ""
    except:
        return False

#* Recipes
def install(recipe, *packages):
    pip = get_pip()
    installed = el.sc_l("{pip} freeze")
    to_install = []
    for p in packages:
        p_v = next((i for i in installed if p + "==" in i), None)
        if p_v:
            print(p_v + ": OK")
        else:
            to_install.append(p)
    if to_install:
        return [lf("{pip} install ") + " ".join(to_install)]
    else:
        return []

def reinstall(recipe, user_input=True):
    dd = el.default_directory()
    git1 = el.locate_dominating_file(dd, ".git")
    git2 = el.file_name_directory(git1)
    pip = get_pip()
    return [
        "cd " + git2,
        lf("{pip} install --upgrade .")]

def sdist(recipe):
    return [
        "rm -rf dist/",
        "python setup.py sdist"]

def clean(recipe):
    return [
        "rm -rf dist",
        "rm -rf *.egg-info"]



#* Imports
import re
import os
import shlex
import pycook.elisp as el
from pycook.elisp import sc, lf
os.environ["TERM"] = "linux"

#* Functions
def install_package(package):
    if type(package) is list:
        [install_package(p) for p in package]
    else:
        res = sc("dpkg --get-selections '{package}'")
        if not len(res) or re.search("deinstall$", res):
            el.bash(lf("sudo apt-get install -y {package}"))
        else:
            print(lf("{package}: OK"))

def git_clone(addr, target, commit=None):
    target = el.expand_file_name(target)
    if el.file_exists_p(target):
        print(lf("{target}: OK"))
    else:
        gdir = el.expand_file_name(target)
        pdir = el.file_name_directory(gdir)
        if not el.file_exists_p(pdir):
            el.make_directory(pdir)
        sc("git clone --recursive {addr} {gdir}")
        if commit:
            sc("cd {pdir} && git reset --hard {commit}")

def symlink_p(fname):
    return " -> " in el.sc("stat {fname}")

def ln(fr, to):
    fr = el.expand_file_name(fr)
    if not el.file_exists_p(fr):
        raise RuntimeError("File doesn't exist", fr)
    to = el.expand_file_name(to)
    if el.file_directory_p(to) and not el.file_directory_p(fr):
        to_full = el.expand_file_name(el.file_name_nondirectory(fr), to)
    else:
        to_full = to
    if el.file_exists_p(to_full):
        if symlink_p(to_full):
            print(lf("{to_full}: OK"))
        else:
            if file_equal(fr, to_full):
                print(lf("{to_full} exists, contents equal"))
            else:
                print(lf("{to_full} exists, contents NOT equal"))
    else:
        fr_abbr = os.path.relpath(fr, os.path.dirname(to))
        el.sc("ln -s {fr_abbr} {to_full}")

def file_equal(f1, f2):
    def md5sum(f):
        return el.sc("md5sum " + shlex.quote(f)).split(" ")[0]
    return md5sum(f1) == md5sum(f2)

def cp(fr, to):
    if el.file_exists_p(to) and file_equal(fr, to):
        print(lf("{to}: OK"))
    else:
        el.sc("cp '{fr}' '{to}'")

def chmod(fname, permissions):
    current = sc("stat -c '%a' {fname}")
    if current == permissions:
        print(lf("{fname}: OK"))
    else:
        if os.access(el.expand_file_name(fname), os.W_OK):
            sc("chmod {permissions} {fname}")
        else:
            el.bash(lf("sudo chmod {permissions} {fname}"))

def barf(fname, text):
    if el.file_exists_p(fname) and text == el.slurp(fname):
        print(lf("{fname}: OK"))
    else:
        quoted_text = shlex.quote(text)
        el.bash(lf("echo {quoted_text} | sudo tee {fname} > /dev/null"))

def make(target, cmds, deps=[]):
    if (el.file_exists_p(target) and
        all([os.path.getctime(target) > os.path.getctime(dep) for dep in deps])):
        print(lf("{target}: OK"))
    else:
        if type(cmds) is str:
            cmds = [cmds]
        elif callable(cmds):
            cmds()
            return
        fcmds = []
        for cmd in cmds:
            cmd1 = re.sub("\\$@", target, cmd)
            cmd2 = re.sub("\\$\\^", " ".join([shlex.quote(dep) for dep in deps]), cmd1)
            cmd3 = re.sub("\\$<", shlex.quote(deps[0]), cmd2) if deps else cmd2
            if el.sc_hookfn:
                el.sc_hookfn(cmd3)
            fcmds.append(cmd3)
        el.bash(fcmds)

def curl(link, directory="~/Software"):
    fname = el.expand_file_name(link.split("/")[-1], directory)
    make(fname, "curl " + link + " -o " + shlex.quote(fname))
    return fname

def render_patch(patch_lines, before):
    regex = "^\\+" if before else "^\\-"
    return "\n".join([line[1:] for line in patch_lines if not re.match(regex, line)])

def patch(fname, patches):
    """Patch FNAME applying PATCHES.
    Each PATCH in PATCHES is in diff -u format.
    Each line in PATCH starts with [+- ].
    [ ] lines are the optional start and end context.
    [-] lines are expected to exist in the file and will be removed.
    [+] lines will be added to the file after [-] lines are removed.

    PATCH lines should be in contiguous order: optional start context,
    [-] lines, [+] lines, optional end context.

    If PATCH was already applied for FNAME, it will be ignored.
    """
    fname = el.expand_file_name(fname)
    fname = os.path.realpath(fname)
    if el.file_exists_p(fname):
        txt = el.slurp(fname)
    else:
        assert not any([
            re.search("^\\-", patch, flags=re.MULTILINE)
            for patch in patches])
        el.sc("touch {fname}")
        txt = ""
    no_change = True
    for patch in patches:
        patch_lines = el.delete("", patch.splitlines())
        chunk_before = render_patch(patch_lines, True)
        chunk_after = render_patch(patch_lines, False)
        if chunk_before == "":
            if chunk_after not in txt:
                no_change = False
                if not (txt == "" or txt[-1] == "\n"):
                    txt += "\n"
                txt += chunk_after + "\n"
        elif chunk_before in txt:
            no_change = False
            txt = txt.replace(chunk_before, chunk_after)
        else:
            # either already patched or fail
            assert chunk_after in txt
    if no_change:
        print(fname + ": OK")
    else:
        el.barf("/tmp/insta.txt", txt)
        cmd = lf("cp /tmp/insta.txt {fname}")
        if not os.access(fname, os.W_OK):
            cmd = "sudo " + cmd
        el.bash(cmd)



#* Imports
import re
import pycook.elisp as el
sc = el.sc

#* Functions
def this_ip():
    out = el.sc("ip route get 1.1.1.1")
    ms = re.findall("src ([0-9.]+)", out)
    assert(len(ms) == 1)
    return ms[0]

#* Recipes
def ip(recipe):
    print(this_ip())

def ip4(recipe):
    # install dnsutils on Debian
    res = el.sc("dig +short myip.opendns.com @resolver1.opendns.com")
    res = res or sc("dig -4 TXT +short o-o.myaddr.l.google.com @ns1.google.com").strip('"')
    print(res)

def ip6(recipe):
    print(sc("dig TXT +short o-o.myaddr.l.google.com @ns1.google.com").strip('"'))



#* Description
__doc__ = "Run an arbitary amount of commands inside Docker."

#* Imports
import sys
import os
import re
import argparse
import json
import pycook.elisp as el
lf = el.lf
dd = el.default_directory()

#* Functions
def docker_opt_mount(fr, to, opts = "rw"):
    return lf("-v {fr}:{to}:{opts}")

def docker_opt_env_mount(v, d):
    return lf("-e {v}={d} ") + docker_opt_mount(d, d)

def docker_opt_tty():
    if sys.stdout.isatty():
        return "-it"
    else:
        return "-i"

def docker_opt_cleanup():
    return "--rm"

def docker_opt_all_ports():
    return "--net=host"

def docker_opt_user():
    return "-e USER=$(id -u) -u $(id -u):$(id -g)"

def docker_opt_display():
    return "-e DISPLAY=$DISPLAY"

def docker_run(image, args, mount = None, env_mount = None, flags=[]):
    cmds_mount = []
    cmds_env_mount = []
    if mount:
        for (f1, f2) in mount:
            cmds_mount += [docker_opt_mount(el.expand_file_name(f1), el.expand_file_name(f2))]
    if env_mount:
        for (d, e) in env_mount:
            el.make_directory(d)
            cmds_env_mount += [docker_opt_env_mount(e, el.expand_file_name(d))]
    cmds = [
        "docker run",
        "-i" if "-T" in flags else docker_opt_tty(),
        docker_opt_cleanup(),
        "" if "-H" in flags else docker_opt_all_ports(),
        docker_opt_user(),
        docker_opt_display(),
        docker_opt_mount(dd, dd),
        "-w " + dd]
    cmds += cmds_env_mount
    cmds += cmds_mount
    cmds += [
        docker_opt_mount("/etc/group", "/etc/group", "ro"),
        docker_opt_mount("/etc/passwd", "/etc/passwd", "ro"),
        docker_opt_mount(dd, dd),
        lf("{image} bash -e -c '{args}'")]
    return " ".join(cmds)

class ArgList:
    def __init__(self):
        self.E = []
        self.m = []
        self.docker_image = None
        self.cmds = []
        self.flags = []

    def __repr__(self):
        return " ".join(
            ["sandbox"] +
            [lf("-E {x[0]} {x[1]}") for x in self.E] +
            [lf("-m {x[0]} {x[1]}") for x in self.m] +
            [self.docker_image or "?"] +
            [self.cmds])

def get_args(argv = None):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("docker_image")
    parser.add_argument("cmds", nargs="?", help="shell arguments", default="bash")
    parser.add_argument(
        "-m",
        action = "append",
        nargs = 2,
        metavar = ("file_fr", "file_to"),
        help = "mount file_fr on host to file_to in docker")
    parser.add_argument(
        "-E",
        action = "append",
        nargs = 2,
        metavar = ("file", "env"),
        help = "mount file on host to itself in docker and point env to it")
    parser.add_argument(
        "-H", help="disable --net=host")
    parser.add_argument(
        "-~", help="mount $HOME")
    if len(argv) <= 1:
        return parser.parse_args(argv)
    args = ArgList()
    xs = list(reversed(argv))
    while xs:
        x = xs.pop()
        if x == "-E":
            args.E += [(xs.pop(), xs.pop())]
        elif x == "-~":
            args.E += [(os.path.expanduser("~"), "HOME")]
        elif x == "-m":
            args.m += [(xs.pop(), xs.pop())]
        elif re.match("-[a-zA-Z]", x):
            args.flags.append(x)
        else:
            args.docker_image = x
            # args.cmds = list(reversed(xs))
            args.cmds = " ".join(reversed(xs))
            break
    return args


def get_image(i):
    if i == "." and el.file_exists_p("params.json"):
        js = json.load(open("params.json", "r"))
        return js["docker_url"] + ":" + js["docker_version"]
    else:
        return i

def main():
    argv = sys.argv[1:]
    if len(argv) == 0:
        argv = ["-h"]
        get_args(argv)
    args = get_args(argv)
    cmd = docker_run(
        get_image(args.docker_image), args.cmds,
        args.m, args.E, args.flags)
    el.bash(cmd)



#* Includes
import sys
import os
import re
import subprocess
import imp
import ast
import inspect
import collections
import pycook.elisp as el
import pycook.recipes as recipes
lf = el.lf

#* Globals
start_dir = el.default_directory()

#* Functions
def recipe_p(x):
    try:
        return inspect.getargspec(x[1]).args[0] == "recipe"
    except:
        return None

def recipe_names_ordered(book):
    body = ast.parse(el.slurp(book)).body
    return [f.name for f in body if isinstance(f, ast.FunctionDef)]

def recipe_dict(book):
    mod = imp.load_source("Cookbook", book)
    funs = inspect.getmembers(mod, inspect.isfunction)
    funs = el.cl_remove_if_not(recipe_p, funs)
    names = recipe_names_ordered(book)
    items = sorted(funs, key = lambda x: el.position(x[0], names, 42))
    return collections.OrderedDict(items)

def book_config(book):
    rc_file = el.expand_file_name("~/.cookrc.py")
    if el.file_exists_p(rc_file):
        mod = imp.load_source("book_config", rc_file)
        config = mod.config
        if book in config:
            return config[book]
        elif "*" in config:
            return config["*"]
    return {}

def recipe_args_description(f):
    arity = recipe_arity(f)
    if arity == 0:
        return ""
    else:
        return " " + " ".join([":" + a for a in function_arglist(f)[1:]])

def recipe_names(book):
    di = recipe_dict(book)
    ns = [k + recipe_args_description(v) for (k, v) in di.items()]
    return "\n".join(ns)

def describe(book, module=""):
    res = "usage: cook"
    if module:
        res += " :" + module
    res += " <recipe>\n"
    res += book + "\n"
    return res + "Available recipes:\n" + recipe_names(book)

def script_get_book():
    dd = el.default_directory()
    d1 = el.locate_dominating_file(dd, "Cookbook.py")
    d2 = el.locate_dominating_file(dd, "cook/Cookbook.py")
    if d1 and (not d2 or len(d1) > len(d2) - 5):
        return (d1, el.file_name_directory(d1))
    elif d2:
        return (d2, el.file_name_directory(el.file_name_directory(d2)))
    else:
        raise RuntimeError("No Cookbook.py or cook/Cookbook.py found")

def get_book():
    caller_frame = sys._getframe().f_back.f_back
    caller_file = caller_frame.f_code.co_filename
    book = os.path.realpath(caller_file)
    return book

def log_file_name(base_dir, book, recipe):
    sub_dir = "_".join(el.delete("", os.path.normpath(book).split(os.sep)[:-1]))
    full_dir = el.expand_file_name(sub_dir, base_dir)
    el.make_directory(full_dir)
    ts = el.replace_regexp_in_string(" ", "_", el.timestamp())
    name = el.lf("{ts}-{recipe}.txt")
    return el.expand_file_name(name, full_dir)

def function_argspec(f):
    try:
        return inspect.getfullargspec(f)
    except:
        return inspect.getargspec(f)

def function_arglist(f):
    spec = function_argspec(f)
    if spec.varargs:
        return [*spec.args, spec.varargs]
    else:
        return spec.args

def recipe_arity(f):
    return len(function_arglist(f)) - 1

def _main(argv, book):
    if len(argv) == 2:
        if argv[1] == "--list":
            print(recipe_names(book))
        elif argv[1] == "--help":
            print(describe(book))
        else:
            recipe = argv[1]
            fun = recipe_dict(book)[recipe]
            cfg = book_config(book)
            if "tee" in cfg and recipe != "bash":
                basedir = cfg["tee"]["location"]
                fname = log_file_name(basedir, book, recipe)
                el.spit(lf("Book: {book}\nRecipe: {recipe}\n"), fname)
                tee = subprocess.Popen(["tee", "-a", fname], stdin = subprocess.PIPE)
                os.dup2(tee.stdin.fileno(), sys.stdout.fileno())
                os.dup2(tee.stdin.fileno(), sys.stderr.fileno())

            cmds = []
            old_sc_hookfn = el.sc_hookfn
            old_cd_hookfn = el.cd_hookfn
            el.sc_hookfn = lambda s: cmds.append("# " + re.sub("\n", "\\\\n", s))
            el.cd_hookfn = lambda d: cmds.append("# cd " + d)
            try:
                ret_cmds = fun(42) or []
            except:
                if cfg.get("pdb", False):
                    import pdb
                    pdb.set_trace()
                    return
                else:
                    raise
            print("\n".join(cmds))
            all_cmds = ret_cmds
            el.sc_hookfn = old_sc_hookfn
            el.cd_hookfn = old_cd_hookfn
            el.bash(all_cmds, echo = True)
    elif len(argv) == 3 and argv[1] == "--pipe":
        mod = imp.load_source("Cookbook", book)
        funs = dict(inspect.getmembers(mod, inspect.isfunction))
        fun = funs[argv[2]]
        os.chdir(start_dir)
        print(fun(input()))
    elif len(argv) >= 2 and recipe_arity(recipe_dict(book)[argv[1]]) == len(argv[2:]):
        recipe = argv[1]
        fun = recipe_dict(book)[recipe]
        el.bash(fun(42, *argv[2:]) or [], echo = True)
    else:
        print(describe(book))

def modules(full = False, match = False):
    cook_dir = el.file_name_directory(recipes.__file__)
    cook_modules = el.directory_files(cook_dir, full, match)
    cook_modules = el.filter(lambda s: not re.search("__", s), cook_modules)
    user_dir = el.expand_file_name("~/.cook.d")
    if el.file_exists_p(user_dir):
        df = el.directory_files(user_dir, full, match)
        user_modules = [
            f for f in df
            if os.path.isfile(el.expand_file_name(f, user_dir))
            and os.path.splitext(f)[1] == ".py"]
    else:
        user_modules = []
    return cook_modules + user_modules

def module_names():
    ms = modules(False, "[^_]\\.py$")
    return el.delete_dups([s[:-3] for s in ms])

def recipe_args(f, args_provided):
    spec = function_argspec(f)
    args_req = spec.args
    assert(args_req[0] == "recipe")
    args_missing = args_req[1 + len(args_provided):]
    args_input = []
    for a in args_missing:
        args_input.append(input(a + ": "))
    return args_provided + args_input

def main(argv = None):
    if argv is None:
        argv = sys.argv
    try:
        if len(argv) >= 2 and re.match("^:", argv[1]):
            module = argv[1][1:]
            if module == "":
                for module in modules(True):
                    print(module)
                sys.exit(0)
            book = get_module(module)
            if len(argv) >= 3:
                recipe = argv[2]
                if recipe == "--list":
                    print(recipe_names(book))
                else:
                    fun = recipe_dict(book)[recipe]
                    cmds = fun(42, *recipe_args(fun, argv[3:])) or []
                    el.bash(cmds)
                    sys.exit(0)
            else:
                print(describe(book, module))
        else:
            (book, dd) = script_get_book()
            os.chdir(dd)
            _main(argv, book)
    except subprocess.CalledProcessError as e:
        # print(e)
        sys.exit(e.returncode)
    except KeyboardInterrupt:
        sys.exit(1)
    except RuntimeError as e:
        print(e)
        sys.exit(1)

def get_module(name):
    """Load a module NAME.
    If two modules are on sys.path, prefer the one on ~/.cook.d/.
    """
    mods = modules(True, name)
    if len(mods) == 2:
        return el.re_filter("\\.cook\\.d/", mods)[0]
    else:
        assert len(mods) == 1, mods
        return mods[0]

def complete(argv = None):
    if argv is None:
        argv = sys.argv
    assert(argv[1] == "cook")
    # below, assume we're completing the last word
    # current word being completed is sys.argv[-1]
    if len(argv) == 3:
        args = argv[2:]
    else:
        args = argv[2:-1]

    # fix the difference between bash-completion.el and the actual bash completion
    if re.match(":.", args[0]):
        args = [":", args[0][1:]] + args[1:]

    if len(args) == 1:
        if args[0] == ":":
            print("\n".join(module_names()))
        else:
            rs = el.sc("cook --list").split("\n")
            # remove extra args
            rs = [re.split(" :", s)[0] for s in rs]
            fr = [r for r in rs if re.match(args[0], r)]
            print("\n".join(fr))
    elif len(args) == 2 and args[0] == ":":
        matching_cands = el.re_filter("^" + args[1], module_names())
        print("\n".join(matching_cands))
    elif len(args) == 3 and args[0] == ":":
        mod = get_module(args[1])
        cands = list(recipe_dict(mod).keys())
        matching_cands = el.re_filter("^" + args[2], cands)
        print("\n".join(matching_cands))
    elif len(args) >= 4 and args[0] == ":":
        mod = get_module(args[1])
        fun = recipe_dict(mod)[args[2]]
        part = args[-1]
        fun_args = function_arglist(fun)
        if len(fun_args) == 2 and fun_args[1] in ["fname", "fnames"]:
            print(el.sc("compgen -f -- {part}"))
        else:
            args = [""]*recipe_arity(fun)
            print(fun(("complete", part), *args))
    else:
        print("")



def list_all_containers(recipe):
    return ["docker ps"]

def list_all_images(recipe):
    return ["docker images"]

def stop_all_containers(recipe):
    return ["docker stop $(docker ps -aq)"]

def remove_stopped_containers(recipe):
    return ["docker rm $(docker ps -aq)"]

def remove_untagged_images(recipe):
    return ["docker rmi $(docker images | grep '^<none>' | awk '{print $3}')"]

def force_service_restart(recipe):
    """Useful when Docker stops working after upgrading kernel."""
    return [
        "sudo rm -rf /var/lib/docker",
        "sudo service docker restart"
    ]

def setup_install_docker(recipe):
    return [
        "sudo apt-get update",
        "sudo apt-get install linux-image-extra-$(uname -r) linux-image-extra-virtual",
        "sudo apt-get update",
        "curl -fsSL https://download.docker.com/linux/ubuntu/gpg | sudo apt-key add -",
        "sudo apt-get install docker-ce",
        "docker --version"]

def setup_run_docker_without_sudo(recipe):
    return [
        "sudo groupadd docker || echo 'ignore'",
        "sudo gpasswd -a $USER docker",
        "echo 'Log out/in to make results premanent'"]



#* Imports
import os
import re
import pycook.elisp as el
import time
from datetime import datetime
sc = el.sc
lf = el.lf

#* Functions
def repo_p(d):
    return el.file_exists_p(el.expand_file_name(".git", d))

def clean_p(repo):
    with el.dd(repo):
        out = sc("git status")
    if (re.search("nothing to commit, working directory clean", out) or
        re.search("nothing added to commit", out)):
        return True
    else:
        return False

def git_time_to_datetime(s):
    t1 = time.strptime(s)
    t2 = time.mktime(t1)
    return datetime.fromtimestamp(t2)

def mtime(repo):
    """Return the last modification time of REPO."""
    if clean_p(repo):
        with el.dd(repo):
            res = git_time_to_datetime(
                el.sc("git log -1 --date=local --format=%cd"))
        return res
    else:
        return datetime.now()

def clone(remote, local):
    res = []
    local = el.expand_file_name(local)
    if el.file_exists_p(local):
        if el.file_exists_p(el.expand_file_name(".git", local)):
            res += [
                "cd " + local,
                "git pull"]
        else:
            raise RuntimeError("Directory exists and is not a git repo", local)
    else:
        (bd, repo) = os.path.split(local)
        el.make_directory(bd)
        res += [
                lf("cd {bd}"),
                lf("git clone {remote} {repo}")]
    return res

#* Recipes
def clean(recipe):
    return ["git reflog expire --expire=now --all && git gc --aggressive --prune=now"]



#* Imports
import sys
import re
import os
import pycook.elisp as el
lf = el.lf

#* Recipes
def compile_and_run(recipe, source_file):
    class_file = re.sub("java$", "class", source_file)
    assert(el.file_exists_p(source_file))
    source = el.file_name_nondirectory(source_file)
    path = el.file_name_directory(source_file)
    name = el.file_name_sans_extension(source)
    res = []
    if (not el.file_exists_p(class_file) or
        el.file_newer_than_file_p(source_file, class_file)):
        res += [lf("javac {source_file}")]
    res += [lf("java -cp {path} {name}")]
    return res



#* Imports
import re
import os
import sys
import pycook.insta as st
import pycook.elisp as el
import pycook.cook as cook
lf = el.lf

#* Functions
def lib_name(fname):
    return "-l:" + el.file_name_nondirectory(fname)

def compile_and_run_cc(inputs, std = "c++11", flags = "", idirs = [], libs = [], cc="g++"):
    """Compile INPUTS together into an executable and run it.

    INPUTS is a list of source files, headers and libraries.
    The executable is named after the first source file.
    Any file that doesn't exist is assumed to be a library.
    """
    if type(inputs) is list:
        main_file = inputs[0]
    else:
        main_file = inputs
    assert(el.file_exists_p(main_file))
    sources = el.re_filter("(cc|cpp|h|hh|hpp)$", inputs)
    diff = set(inputs) - set(sources)
    libs_inl = [x for x in inputs if x in diff]
    libs_str = " " + " ".join(["-l" + lib for lib in libs_inl])
    lib_dirs = set(el.mapcar(el.file_name_directory, libs))
    lflags = el.mapcar(lib_name, libs)
    libs_str += " " + " ".join(["-L" + d for d in lib_dirs] + lflags)
    exe_file = re.sub("cc$", "e", main_file)
    res = []
    flags += " " + " ".join(["-I" + d for d in idirs])
    if (not el.file_exists_p(exe_file) or
        any([el.file_newer_than_file_p(f, exe_file) for f in sources + [cook.script_get_book()[0]]])):
        ccmd = lf("g++ -g -O2 -std={std} {flags} -o {exe_file} ") + " ".join(sources) + libs_str
        res += [ccmd]
    if lib_dirs:
        res += ["export LD_LIBRARY_PATH=" + ":".join(lib_dirs)]

    res += [lf("./{exe_file}")]
    return res

def compile_and_run(fname):
    (base, ext) = os.path.splitext(fname)
    if ext == ".c":
        return compile_and_run_cc([fname], cc="gcc", std="c99")
    elif ext in [".cc", ".cpp"]:
        return compile_and_run_cc([fname])
    else:
        raise RuntimeError("Unexpected file type", fname)



#* Imports
import pycook.elisp as el
lf = el.lf

#* Functions
def compile_and_run(src_file):
    exe_file = el.file_name_sans_extension(src_file)
    if (el.file_exists_p(exe_file) and
        el.file_newer_than_file_p(exe_file, src_file)):
        return [lf("./{exe_file}")]
    else:
        return [
            lf("rustc {src_file}"),
            lf("./{exe_file}")]



from setuptools import setup, find_packages

setup(
    name='pycook',
    version='0.7.38',
    description='Makefile -> Cookbook.py.',
    long_description='https://raw.githubusercontent.com/abo-abo/cook/master/README.org',
    packages=find_packages(exclude=['etc', 'examples']),
    url='https://github.com/abo-abo/cook',
    author='Oleh Krehel',
    author_email='ohwoeowho@gmail.com',
    license='GPLv3+',
    keywords='makefile automation',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'],
    install_requires=[],
    data_files=[('cook', ['README', 'etc/bash-completion.sh', 'cook.el', 'etc/scripts.el', 'etc/elpa.el'])],
    scripts=['etc/cook', 'etc/_cook_complete'],
    entry_points={'console_scripts': [
        'sandbox=pycook.sandbox:main'
    ]}
)



#* Imports
import os
import pycook.elisp as el
import pycook.insta as st

#* Recipes
def update(recipe):
    cook_el = el.emacs_cook_script("cook.el")
    user_p = " --user " if "/.local/" in cook_el else " "
    el.bash(el.lf("pip3 install{user_p}--upgrade pycook"))
    if "INSIDE_EMACS" in os.environ:
        el.sc(el.emacsclient_eval(el.lf('(load-file "{cook_el}")')))

def setup_dotfiles(recipe):
    st.patch("~/.profile", ['+PATH="$HOME/.local/bin:$PATH"'])
    st.patch("~/.bashrc", ["""
+if [ -f ~/.local/cook/bash-completion.sh ]; then
+    . ~/.local/cook/bash-completion.sh
+fi
"""])
    st.patch("~/.bash_profile", ["+. ~/.profile", "+. ~/.bashrc"])

def profile(recipe):
    return "pyinstrument $(which cook) :linux ls_users"
